"use strict";
var __decorate = (this && this.__decorate) || function (decorators, target, key, desc) {
    var c = arguments.length, r = c < 3 ? target : desc === null ? desc = Object.getOwnPropertyDescriptor(target, key) : desc, d;
    if (typeof Reflect === "object" && typeof Reflect.decorate === "function") r = Reflect.decorate(decorators, target, key, desc);
    else for (var i = decorators.length - 1; i >= 0; i--) if (d = decorators[i]) r = (c < 3 ? d(r) : c > 3 ? d(target, key, r) : d(target, key)) || r;
    return c > 3 && r && Object.defineProperty(target, key, r), r;
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.DifPresentationExchangeService = void 0;
const pex_1 = require("@sphereon/pex");
const tsyringe_1 = require("tsyringe");
const crypto_1 = require("../../crypto");
const error_1 = require("../../error");
const utils_1 = require("../../utils");
const dids_1 = require("../dids");
const vc_1 = require("../vc");
const DifPresentationExchangeError_1 = require("./DifPresentationExchangeError");
const models_1 = require("./models");
const utils_2 = require("./utils");
let DifPresentationExchangeService = class DifPresentationExchangeService {
    constructor() {
        this.pex = new pex_1.PEX();
    }
    async getCredentialsForRequest(agentContext, presentationDefinition) {
        const credentialRecords = await this.queryCredentialForPresentationDefinition(agentContext, presentationDefinition);
        // FIXME: why are we resolving all created dids here?
        // If we want to do this we should extract all dids from the credential records and only
        // fetch the dids for the queried credential records
        const didsApi = agentContext.dependencyManager.resolve(dids_1.DidsApi);
        const didRecords = await didsApi.getCreatedDids();
        const holderDids = didRecords.map((didRecord) => didRecord.did);
        return (0, utils_2.getCredentialsForRequest)(presentationDefinition, credentialRecords, holderDids);
    }
    /**
     * Selects the credentials to use based on the output from `getCredentialsForRequest`
     * Use this method if you don't want to manually select the credentials yourself.
     */
    selectCredentialsForRequest(credentialsForRequest) {
        if (!credentialsForRequest.areRequirementsSatisfied) {
            throw new error_1.AriesFrameworkError('Could not find the required credentials for the presentation submission');
        }
        const credentials = {};
        for (const requirement of credentialsForRequest.requirements) {
            for (const submission of requirement.submissionEntry) {
                if (!credentials[submission.inputDescriptorId]) {
                    credentials[submission.inputDescriptorId] = [];
                }
                // We pick the first matching VC if we are auto-selecting
                credentials[submission.inputDescriptorId].push(submission.verifiableCredentials[0].credential);
            }
        }
        return credentials;
    }
    validatePresentationDefinition(presentationDefinition) {
        const validation = pex_1.PEX.validateDefinition(presentationDefinition);
        const errorMessages = this.formatValidated(validation);
        if (errorMessages.length > 0) {
            throw new DifPresentationExchangeError_1.DifPresentationExchangeError(`Invalid presentation definition`, { additionalMessages: errorMessages });
        }
    }
    validatePresentationSubmission(presentationSubmission) {
        const validation = pex_1.PEX.validateSubmission(presentationSubmission);
        const errorMessages = this.formatValidated(validation);
        if (errorMessages.length > 0) {
            throw new DifPresentationExchangeError_1.DifPresentationExchangeError(`Invalid presentation submission`, { additionalMessages: errorMessages });
        }
    }
    validatePresentation(presentationDefinition, presentation) {
        const { errors } = this.pex.evaluatePresentation(presentationDefinition, presentation.encoded);
        if (errors) {
            const errorMessages = this.formatValidated(errors);
            if (errorMessages.length > 0) {
                throw new DifPresentationExchangeError_1.DifPresentationExchangeError(`Invalid presentation`, { additionalMessages: errorMessages });
            }
        }
    }
    formatValidated(v) {
        const validated = Array.isArray(v) ? v : [v];
        return validated
            .filter((r) => r.tag === pex_1.Status.ERROR)
            .map((r) => r.message)
            .filter((r) => Boolean(r));
    }
    /**
     * Queries the wallet for credentials that match the given presentation definition. This only does an initial query based on the
     * schema of the input descriptors. It does not do any further filtering based on the constraints in the input descriptors.
     */
    async queryCredentialForPresentationDefinition(agentContext, presentationDefinition) {
        const w3cCredentialRepository = agentContext.dependencyManager.resolve(vc_1.W3cCredentialRepository);
        const query = [];
        const presentationDefinitionVersion = pex_1.PEX.definitionVersionDiscovery(presentationDefinition);
        if (!presentationDefinitionVersion.version) {
            throw new DifPresentationExchangeError_1.DifPresentationExchangeError(`Unable to determine the Presentation Exchange version from the presentation definition
        `, presentationDefinitionVersion.error ? { additionalMessages: [presentationDefinitionVersion.error] } : {});
        }
        if (presentationDefinitionVersion.version === pex_1.PEVersion.v1) {
            const pd = presentationDefinition;
            // The schema.uri can contain either an expanded type, or a context uri
            for (const inputDescriptor of pd.input_descriptors) {
                for (const schema of inputDescriptor.schema) {
                    query.push({
                        $or: [{ expandedType: [schema.uri] }, { contexts: [schema.uri] }, { type: [schema.uri] }],
                    });
                }
            }
        }
        else if (presentationDefinitionVersion.version === pex_1.PEVersion.v2) {
            // FIXME: As PE version 2 does not have the `schema` anymore, we can't query by schema anymore.
            // For now we retrieve ALL credentials, as we did the same for V1 with JWT credentials. We probably need
            // to find some way to do initial filtering, hopefully if there's a filter on the `type` field or something.
        }
        else {
            throw new DifPresentationExchangeError_1.DifPresentationExchangeError(`Unsupported presentation definition version ${presentationDefinitionVersion.version}`);
        }
        // query the wallet ourselves first to avoid the need to query the pex library for all
        // credentials for every proof request
        const credentialRecords = query.length > 0
            ? await w3cCredentialRepository.findByQuery(agentContext, {
                $or: query,
            })
            : await w3cCredentialRepository.getAll(agentContext);
        return credentialRecords;
    }
    addCredentialToSubjectInputDescriptor(subjectsToInputDescriptors, subjectId, inputDescriptorId, credential) {
        var _a, _b;
        const inputDescriptorsToCredentials = (_a = subjectsToInputDescriptors[subjectId]) !== null && _a !== void 0 ? _a : {};
        const credentials = (_b = inputDescriptorsToCredentials[inputDescriptorId]) !== null && _b !== void 0 ? _b : [];
        credentials.push(credential);
        inputDescriptorsToCredentials[inputDescriptorId] = credentials;
        subjectsToInputDescriptors[subjectId] = inputDescriptorsToCredentials;
    }
    getPresentationFormat(presentationDefinition, credentials) {
        const allCredentialsAreJwtVc = credentials === null || credentials === void 0 ? void 0 : credentials.every((c) => typeof c === 'string');
        const allCredentialsAreLdpVc = credentials === null || credentials === void 0 ? void 0 : credentials.every((c) => typeof c !== 'string');
        const inputDescriptorsNotSupportingJwtVc = presentationDefinition.input_descriptors.filter((d) => d.format && d.format.jwt_vc === undefined);
        const inputDescriptorsNotSupportingLdpVc = presentationDefinition.input_descriptors.filter((d) => d.format && d.format.ldp_vc === undefined);
        if (allCredentialsAreJwtVc &&
            (presentationDefinition.format === undefined || presentationDefinition.format.jwt_vc) &&
            inputDescriptorsNotSupportingJwtVc.length === 0) {
            return vc_1.ClaimFormat.JwtVp;
        }
        else if (allCredentialsAreLdpVc &&
            (presentationDefinition.format === undefined || presentationDefinition.format.ldp_vc) &&
            inputDescriptorsNotSupportingLdpVc.length === 0) {
            return vc_1.ClaimFormat.LdpVp;
        }
        else {
            throw new DifPresentationExchangeError_1.DifPresentationExchangeError('No suitable presentation format found for the given presentation definition, and credentials');
        }
    }
    async createPresentation(agentContext, options) {
        const { presentationDefinition, challenge, nonce, domain, presentationSubmissionLocation } = options;
        const proofStructure = {};
        Object.entries(options.credentialsForInputDescriptor).forEach(([inputDescriptorId, credentials]) => {
            credentials.forEach((credential) => {
                const subjectId = credential.credentialSubjectIds[0];
                if (!subjectId) {
                    throw new DifPresentationExchangeError_1.DifPresentationExchangeError('Missing required credential subject for creating the presentation.');
                }
                this.addCredentialToSubjectInputDescriptor(proofStructure, subjectId, inputDescriptorId, credential);
            });
        });
        const verifiablePresentationResultsWithFormat = [];
        const subjectToInputDescriptors = Object.entries(proofStructure);
        for (const [subjectId, subjectInputDescriptorsToCredentials] of subjectToInputDescriptors) {
            // Determine a suitable verification method for the presentation
            const verificationMethod = await this.getVerificationMethodForSubjectId(agentContext, subjectId);
            if (!verificationMethod) {
                throw new DifPresentationExchangeError_1.DifPresentationExchangeError(`No verification method found for subject id '${subjectId}'.`);
            }
            // We create a presentation for each subject
            // Thus for each subject we need to filter all the related input descriptors and credentials
            // FIXME: cast to V1, as tsc errors for strange reasons if not
            const inputDescriptorsForSubject = presentationDefinition.input_descriptors.filter((inputDescriptor) => inputDescriptor.id in subjectInputDescriptorsToCredentials);
            // Get all the credentials associated with the input descriptors
            const credentialsForSubject = Object.values(subjectInputDescriptorsToCredentials)
                .flat()
                .map(utils_2.getSphereonOriginalVerifiableCredential);
            const presentationDefinitionForSubject = Object.assign(Object.assign({}, presentationDefinition), { input_descriptors: inputDescriptorsForSubject, 
                // We remove the submission requirements, as it will otherwise fail to create the VP
                submission_requirements: undefined });
            const format = this.getPresentationFormat(presentationDefinitionForSubject, credentialsForSubject);
            // FIXME: Q1: is holder always subject id, what if there are multiple subjects???
            // FIXME: Q2: What about proofType, proofPurpose verification method for multiple subjects?
            const verifiablePresentationResult = await this.pex.verifiablePresentationFrom(presentationDefinitionForSubject, credentialsForSubject, this.getPresentationSignCallback(agentContext, verificationMethod, format), {
                holderDID: subjectId,
                proofOptions: { challenge, domain, nonce },
                signatureOptions: { verificationMethod: verificationMethod === null || verificationMethod === void 0 ? void 0 : verificationMethod.id },
                presentationSubmissionLocation: presentationSubmissionLocation !== null && presentationSubmissionLocation !== void 0 ? presentationSubmissionLocation : models_1.DifPresentationExchangeSubmissionLocation.PRESENTATION,
            });
            verifiablePresentationResultsWithFormat.push({ verifiablePresentationResult, format });
        }
        if (!verifiablePresentationResultsWithFormat[0]) {
            throw new DifPresentationExchangeError_1.DifPresentationExchangeError('No verifiable presentations created');
        }
        if (subjectToInputDescriptors.length !== verifiablePresentationResultsWithFormat.length) {
            throw new DifPresentationExchangeError_1.DifPresentationExchangeError('Invalid amount of verifiable presentations created');
        }
        const presentationSubmission = {
            id: verifiablePresentationResultsWithFormat[0].verifiablePresentationResult.presentationSubmission.id,
            definition_id: verifiablePresentationResultsWithFormat[0].verifiablePresentationResult.presentationSubmission.definition_id,
            descriptor_map: [],
        };
        for (const vpf of verifiablePresentationResultsWithFormat) {
            const { verifiablePresentationResult } = vpf;
            presentationSubmission.descriptor_map.push(...verifiablePresentationResult.presentationSubmission.descriptor_map);
        }
        return {
            verifiablePresentations: verifiablePresentationResultsWithFormat.map((r) => (0, utils_2.getW3cVerifiablePresentationInstance)(r.verifiablePresentationResult.verifiablePresentation)),
            presentationSubmission,
            presentationSubmissionLocation: verifiablePresentationResultsWithFormat[0].verifiablePresentationResult.presentationSubmissionLocation,
        };
    }
    getSigningAlgorithmFromVerificationMethod(verificationMethod, suitableAlgorithms) {
        const key = (0, dids_1.getKeyFromVerificationMethod)(verificationMethod);
        const jwk = (0, crypto_1.getJwkFromKey)(key);
        if (suitableAlgorithms) {
            const possibleAlgorithms = jwk.supportedSignatureAlgorithms.filter((alg) => suitableAlgorithms === null || suitableAlgorithms === void 0 ? void 0 : suitableAlgorithms.includes(alg));
            if (!possibleAlgorithms || possibleAlgorithms.length === 0) {
                throw new DifPresentationExchangeError_1.DifPresentationExchangeError([
                    `Found no suitable signing algorithm.`,
                    `Algorithms supported by Verification method: ${jwk.supportedSignatureAlgorithms.join(', ')}`,
                    `Suitable algorithms: ${suitableAlgorithms.join(', ')}`,
                ].join('\n'));
            }
        }
        const alg = jwk.supportedSignatureAlgorithms[0];
        if (!alg)
            throw new DifPresentationExchangeError_1.DifPresentationExchangeError(`No supported algs for key type: ${key.keyType}`);
        return alg;
    }
    getSigningAlgorithmsForPresentationDefinitionAndInputDescriptors(algorithmsSatisfyingDefinition, inputDescriptorAlgorithms) {
        const allDescriptorAlgorithms = inputDescriptorAlgorithms.flat();
        const algorithmsSatisfyingDescriptors = allDescriptorAlgorithms.filter((alg) => inputDescriptorAlgorithms.every((descriptorAlgorithmSet) => descriptorAlgorithmSet.includes(alg)));
        const algorithmsSatisfyingPdAndDescriptorRestrictions = algorithmsSatisfyingDefinition.filter((alg) => algorithmsSatisfyingDescriptors.includes(alg));
        if (algorithmsSatisfyingDefinition.length > 0 &&
            algorithmsSatisfyingDescriptors.length > 0 &&
            algorithmsSatisfyingPdAndDescriptorRestrictions.length === 0) {
            throw new DifPresentationExchangeError_1.DifPresentationExchangeError(`No signature algorithm found for satisfying restrictions of the presentation definition and input descriptors`);
        }
        if (allDescriptorAlgorithms.length > 0 && algorithmsSatisfyingDescriptors.length === 0) {
            throw new DifPresentationExchangeError_1.DifPresentationExchangeError(`No signature algorithm found for satisfying restrictions of the input descriptors`);
        }
        let suitableAlgorithms;
        if (algorithmsSatisfyingPdAndDescriptorRestrictions.length > 0) {
            suitableAlgorithms = algorithmsSatisfyingPdAndDescriptorRestrictions;
        }
        else if (algorithmsSatisfyingDescriptors.length > 0) {
            suitableAlgorithms = algorithmsSatisfyingDescriptors;
        }
        else if (algorithmsSatisfyingDefinition.length > 0) {
            suitableAlgorithms = algorithmsSatisfyingDefinition;
        }
        return suitableAlgorithms;
    }
    getSigningAlgorithmForJwtVc(presentationDefinition, verificationMethod) {
        var _a, _b, _c;
        const algorithmsSatisfyingDefinition = (_c = (_b = (_a = presentationDefinition.format) === null || _a === void 0 ? void 0 : _a.jwt_vc) === null || _b === void 0 ? void 0 : _b.alg) !== null && _c !== void 0 ? _c : [];
        const inputDescriptorAlgorithms = presentationDefinition.input_descriptors
            .map((descriptor) => { var _a, _b, _c; return (_c = (_b = (_a = descriptor.format) === null || _a === void 0 ? void 0 : _a.jwt_vc) === null || _b === void 0 ? void 0 : _b.alg) !== null && _c !== void 0 ? _c : []; })
            .filter((alg) => alg.length > 0);
        const suitableAlgorithms = this.getSigningAlgorithmsForPresentationDefinitionAndInputDescriptors(algorithmsSatisfyingDefinition, inputDescriptorAlgorithms);
        return this.getSigningAlgorithmFromVerificationMethod(verificationMethod, suitableAlgorithms);
    }
    getProofTypeForLdpVc(agentContext, presentationDefinition, verificationMethod) {
        var _a, _b, _c;
        const algorithmsSatisfyingDefinition = (_c = (_b = (_a = presentationDefinition.format) === null || _a === void 0 ? void 0 : _a.ldp_vc) === null || _b === void 0 ? void 0 : _b.proof_type) !== null && _c !== void 0 ? _c : [];
        const inputDescriptorAlgorithms = presentationDefinition.input_descriptors
            .map((descriptor) => { var _a, _b, _c; return (_c = (_b = (_a = descriptor.format) === null || _a === void 0 ? void 0 : _a.ldp_vc) === null || _b === void 0 ? void 0 : _b.proof_type) !== null && _c !== void 0 ? _c : []; })
            .filter((alg) => alg.length > 0);
        const suitableSignatureSuites = this.getSigningAlgorithmsForPresentationDefinitionAndInputDescriptors(algorithmsSatisfyingDefinition, inputDescriptorAlgorithms);
        // For each of the supported algs, find the key types, then find the proof types
        const signatureSuiteRegistry = agentContext.dependencyManager.resolve(vc_1.SignatureSuiteRegistry);
        const supportedSignatureSuite = signatureSuiteRegistry.getByVerificationMethodType(verificationMethod.type);
        if (!supportedSignatureSuite) {
            throw new DifPresentationExchangeError_1.DifPresentationExchangeError(`Couldn't find a supported signature suite for the given verification method type '${verificationMethod.type}'`);
        }
        if (suitableSignatureSuites) {
            if (suitableSignatureSuites.includes(supportedSignatureSuite.proofType) === false) {
                throw new DifPresentationExchangeError_1.DifPresentationExchangeError([
                    'No possible signature suite found for the given verification method.',
                    `Verification method type: ${verificationMethod.type}`,
                    `SupportedSignatureSuite '${supportedSignatureSuite.proofType}'`,
                    `SuitableSignatureSuites: ${suitableSignatureSuites.join(', ')}`,
                ].join('\n'));
            }
            return supportedSignatureSuite.proofType;
        }
        return supportedSignatureSuite.proofType;
    }
    getPresentationSignCallback(agentContext, verificationMethod, vpFormat) {
        const w3cCredentialService = agentContext.dependencyManager.resolve(vc_1.W3cCredentialService);
        return async (callBackParams) => {
            var _a, _b, _c, _d;
            // The created partial proof and presentation, as well as original supplied options
            const { presentation: presentationJson, options, presentationDefinition } = callBackParams;
            const { challenge, domain, nonce } = (_a = options.proofOptions) !== null && _a !== void 0 ? _a : {};
            const { verificationMethod: verificationMethodId } = (_b = options.signatureOptions) !== null && _b !== void 0 ? _b : {};
            if (verificationMethodId && verificationMethodId !== verificationMethod.id) {
                throw new DifPresentationExchangeError_1.DifPresentationExchangeError(`Verification method from signing options ${verificationMethodId} does not match verification method ${verificationMethod.id}`);
            }
            let signedPresentation;
            if (vpFormat === 'jwt_vp') {
                signedPresentation = await w3cCredentialService.signPresentation(agentContext, {
                    format: vc_1.ClaimFormat.JwtVp,
                    alg: this.getSigningAlgorithmForJwtVc(presentationDefinition, verificationMethod),
                    verificationMethod: verificationMethod.id,
                    presentation: utils_1.JsonTransformer.fromJSON(presentationJson, vc_1.W3cPresentation),
                    challenge: (_c = challenge !== null && challenge !== void 0 ? challenge : nonce) !== null && _c !== void 0 ? _c : (await agentContext.wallet.generateNonce()),
                    domain,
                });
            }
            else if (vpFormat === 'ldp_vp') {
                signedPresentation = await w3cCredentialService.signPresentation(agentContext, {
                    format: vc_1.ClaimFormat.LdpVp,
                    proofType: this.getProofTypeForLdpVc(agentContext, presentationDefinition, verificationMethod),
                    proofPurpose: 'authentication',
                    verificationMethod: verificationMethod.id,
                    presentation: utils_1.JsonTransformer.fromJSON(presentationJson, vc_1.W3cPresentation),
                    challenge: (_d = challenge !== null && challenge !== void 0 ? challenge : nonce) !== null && _d !== void 0 ? _d : (await agentContext.wallet.generateNonce()),
                    domain,
                });
            }
            else {
                throw new DifPresentationExchangeError_1.DifPresentationExchangeError(`Only JWT credentials or JSONLD credentials are supported for a single presentation`);
            }
            return (0, utils_2.getSphereonW3cVerifiablePresentation)(signedPresentation);
        };
    }
    async getVerificationMethodForSubjectId(agentContext, subjectId) {
        const didsApi = agentContext.dependencyManager.resolve(dids_1.DidsApi);
        if (!subjectId.startsWith('did:')) {
            throw new DifPresentationExchangeError_1.DifPresentationExchangeError(`Only dids are supported as credentialSubject id. ${subjectId} is not a valid did`);
        }
        const didDocument = await didsApi.resolveDidDocument(subjectId);
        if (!didDocument.authentication || didDocument.authentication.length === 0) {
            throw new DifPresentationExchangeError_1.DifPresentationExchangeError(`No authentication verificationMethods found for did ${subjectId} in did document`);
        }
        // the signature suite to use for the presentation is dependant on the credentials we share.
        // 1. Get the verification method for this given proof purpose in this DID document
        let [verificationMethod] = didDocument.authentication;
        if (typeof verificationMethod === 'string') {
            verificationMethod = didDocument.dereferenceKey(verificationMethod, ['authentication']);
        }
        return verificationMethod;
    }
};
DifPresentationExchangeService = __decorate([
    (0, tsyringe_1.injectable)()
], DifPresentationExchangeService);
exports.DifPresentationExchangeService = DifPresentationExchangeService;
//# sourceMappingURL=DifPresentationExchangeService.js.map